	.include "ATxmega32A4Udef.inc"

#if PROGMEM_SIZE>$10000
.equ big_flash=1
#else
.equ big_flash=0
#endif

.equ PORTA=VPORT0_OUT
.equ PINA=VPORT0_IN
.equ DDRA=VPORT0_DIR
.equ PORTB=VPORT1_OUT
.equ PINB=VPORT1_IN
.equ DDRB=VPORT1_DIR
.equ PORTC=VPORT2_OUT
.equ PINC=VPORT2_IN
.equ DDRC=VPORT2_DIR
.equ PORTD=VPORT3_OUT
.equ PIND=VPORT3_IN
.equ DDRD=VPORT3_DIR

;select one out of the following three
.equ flash_rd=1
.equ usart_rd=0
.equ spiram_rd=0								;23LC1024, 128 Kbytes serial RAM

.equ lcd_dma=1

;#define RAM_USART								;use USART in master SPI mode for external RAM interface
#define RAM_SPI									;use native SPI for external RAM interface

;#define LCD_CTRL_ILI9341						;for 320x240 TFT
#define LCD_CTRL_ST7735R						;for 160x128 TFT

;Pin definitions for LCD interface, all pins should be on the same port
#define DISPLAYPORT		PORTC
#define DISPLAYPORT_dd	DDRC
#define DISPLAY_RESET	0
#define DISPLAY_C_D		1
#define DISPLAY_SS		4
#define DISPLAY_SCK		5
#define DISPLAY_MOSI	7
#define DISPLAY_DMA_TRIG DMA_CH_TRIGSRC_USARTC1_DRE_gc
.equ    DISPLAY_IF	   =USARTC1_base

;Pin definitions for interface to external RAM, all pins should be on the same port
#ifdef RAM_USART
#define RAMPORT 		PORTD
#define RAMPORT_dd 		DDRD
#define RAMPORT_SS 		0
#define RAMPORT_SCK 	1
#define RAMPORT_MOSI	3
.equ    RAM_IF		   =USARTD0_base
.equ	RAM_IF_MODE    =1
#endif

#ifdef RAM_SPI
#define RAMPORT 		PORTD
#define RAMPORT_dd 		DDRD
#define RAMPORT_SS 		4
#define RAMPORT_MOSI	5
#define RAMPORT_SCK 	7
.equ	RAM_IF		   =SPID_base
.equ	RAM_IF_MODE	   =0
#endif

;Pin definition for selection of serial RAM loader
#define MODE_PIN_PORT	PORTE_IN
#define MODE_PIN		0						;put low level to pin and trigger reset to enter loader
#define MODE_PIN_CTRL	PORTE_PIN0CTRL

;Pin definitions for serial interface
#define SERIALPORT		PORTC
#define SERIALPORT_dd	DDRC
#define SERIALPORT_TX	3
.equ    SERIAL_IF	   =USARTC0_base
.equ    SERIAL_IF_INTV =USARTC0_vbase

;Pin definition for RTS control output of serial interface
#define RTS 			PORTA,0
#define RTS_dd 			DDRA,0

.equ	DELAY_TMR	   =TCC0_base
.equ	DELAY_TMR_INTV =TCC0_OVF_vect

;image size is hard coded
;choose both dimensions in steps of 16 pixels
.equ x_=128	;240
.equ y_=160	;320

;CPU clock, choose in steps of 8 MHz
.equ clock=32;MHz

	.if usart_rd==1 || spiram_rd==1

.equ bitrate=460800;bit/s						;use 115200, 230400, 460800 or any other bitrate down to 19200 bit/s

.set bsel=((clock*62500*128)/bitrate)-128
.set bscale=-7
	.if bsel>4095
.set bsel=((clock*62500*64)/bitrate)-64
.set bscale=-6
	.endif
	.if bsel>4095
.set bsel=((clock*62500*32)/bitrate)-32
.set bscale=-5
	.endif
	.if bsel>4095
.set bsel=((clock*62500*16)/bitrate)-16
.set bscale=-4
	.endif

	.endif

	.org 0
	rjmp start

;delay timer interrupt
	.org DELAY_TMR_INTV
	push r16
	in r16,CPU_SREG
	push r16
	lds r16,(tmcnt)
	tst r16
	breq timerx
	dec r16
	sts (tmcnt),r16
timerx: pop r16
	out CPU_SREG,r16
	pop r16
	reti

	.if usart_rd
;serial interface receive interrupt
;writes received bytes into ring buffer
;handles RTS if buffer's getting full
	.org SERIAL_IF_INTV+USART_RXC_voffset
	push ZL
	push ZH
	push r16
	push r0
	in r0,CPU_SREG
	clr ZH
	lds ZL,(wr_sptr)
	subi ZL,LOW(-serbuf)
	sbci ZH,HIGH(-serbuf)
	lds r16,(SERIAL_IF+USART_DATA_offset)
	st Z+,r16
	subi ZL,LOW(serbuf)
	sts (wr_sptr),ZL
	lds r16,(rd_sptr)
	sub ZL,r16
	cpi ZL,248
	brcs rx_intx
	sbi RTS										;RX buffer is getting full, disable RTS
rx_intx: out CPU_SREG,r0
	pop r0
	pop r16
	pop ZH
	pop ZL
	reti
	.endif

start:
;assign PORT A, B, C and D to virtual ports in ATMega manner
	ldi r16,PORTCFG_VP13MAP_PORTB_gc|PORTCFG_VP02MAP_PORTA_gc
	sts (PORTCFG_VPCTRLA),r16
	ldi r16,PORTCFG_VP13MAP_PORTD_gc|PORTCFG_VP02MAP_PORTC_gc
	sts (PORTCFG_VPCTRLB),r16
	
	.if spiram_rd
	ldi r16,PORT_OPC_PULLUP_gc
	sts (MODE_PIN_CTRL),r16
	.endif

	ldi r17,CCP_IOREG_gc
	out CPU_CCP,r17
	ldi r16,OSC_RC32MEN_bm						;enable 32Mhz internal clock
	sts (OSC_CTRL),r16
wait_for_osc: lds r16,(OSC_STATUS)
	andi r16,OSC_RC32MRDY_bm
	breq wait_for_osc

	ldi r16,1
	sts (DFLLRC32M_CTRL),r16
	ldi r16,OSC_PLLSRC_RC32M_gc+clock/8			;Source internal 32MHz/4
	sts (OSC_PLLCTRL),r16

	out CPU_CCP,r17
	ldi r16,OSC_PLLEN_bm+OSC_RC32MEN_bm			;enable PLL + 32MHz internal clock
	sts (OSC_CTRL),r16
wait_for_osc1: lds r16,(OSC_STATUS)
	andi r16,OSC_PLLRDY_bm
	breq wait_for_osc1	

	out CPU_CCP,r17
	ldi r16,CLK_SCLKSEL_PLL_gc					;select PLL as clock source
	sts (CLK_CTRL),r16

;display reset to H and output
;display C/D to H and output
;display port /SS of SPI to H and output
;display port SCK of SPI to H and output
;display port MOSI of SPI to H and output
	in r16,DISPLAYPORT
	ori r16,1<<DISPLAY_MOSI|1<<DISPLAY_SCK|1<<DISPLAY_SS|1<<DISPLAY_C_D|1<<DISPLAY_RESET
	out DISPLAYPORT,r16
	in r16,DISPLAYPORT_dd
	ori r16,1<<DISPLAY_MOSI|1<<DISPLAY_SCK|1<<DISPLAY_SS|1<<DISPLAY_C_D|1<<DISPLAY_RESET
	out DISPLAYPORT_dd,r16

	.if usart_rd==1 || spiram_rd==1
	sbi SERIALPORT,SERIALPORT_TX				;Tx of USART to H and output
	sbi SERIALPORT_dd,SERIALPORT_TX
	.endif
	.if usart_rd==1
	sbi RTS_dd									;RTS to output
	.endif

	.if spiram_rd==1
;/SS of SPI to H and output
;SCK of SPI to H and output
;MOSI of SPI to H and output
	in r16,RAMPORT
	ori r16,1<<RAMPORT_MOSI|1<<RAMPORT_SCK|1<<RAMPORT_SS
	out RAMPORT,r16
	in r16,RAMPORT_dd
	ori r16,1<<RAMPORT_MOSI|1<<RAMPORT_SCK|1<<RAMPORT_SS
	out RAMPORT_dd,r16
	.endif

	ldi r16,PMIC_HILVLEN_bm|PMIC_MEDLVLEN_bm|PMIC_LOLVLEN_bm	;enable all interrupt levels
	sts (PMIC_CTRL),r16

	.if usart_rd==1
;initialize serial interface
	ldi YL,LOW(SERIAL_IF)
	ldi YH,HIGH(SERIAL_IF)
	ldi r16,LOW(bsel)
	std Y+USART_BAUDCTRLA_offset,r16
	ldi r16,bscale<<4|HIGH(bsel)
	std Y+USART_BAUDCTRLB_offset,r16
	ldi r16,USART_CMODE_ASYNCHRONOUS_gc|USART_CHSIZE_8BIT_gc	;async 8n1
	std Y+USART_CTRLC_offset,r16
	ldi r16,USART_RXCINTLVL_LO_gc				;low rx interrupt priority
	std Y+USART_CTRLA_offset,r16
	ldi r16,USART_RXEN_bm|USART_TXEN_bm
	std Y+USART_CTRLB_offset,r16
	ldi r16,USART_RXCIF_bm
	std Y+USART_STATUS_offset,r16
	.endif

;initialize delay timer
	ldi YL,LOW(DELAY_TMR)
	ldi YH,HIGH(DELAY_TMR)
	ldi r16,LOW(clock/8*1000)
	std Y+TC0_PER_offset,r16
	ldi r16,HIGH(clock/8*1000)
	std Y+TC0_PER_offset+1,r16
	ldi r16,TC_CLKSEL_DIV8_gc
	std Y+TC0_CTRLA_offset,r16
	ldi r16,TC_OVFINTLVL_LO_gc					;low timer interrupt priority
	std Y+TC0_INTCTRLA_offset,r16
	sei

	cbi DISPLAYPORT,DISPLAY_RESET
	ldi r16,2									;assert LCD reset for 2 milliseconds
	rcall wait
	sbi DISPLAYPORT,DISPLAY_RESET
	ldi r16,125									;wait 125 milliseconds after LCD reset
	rcall wait

;initialize display interface, use USART in master SPI mode to enable DMA usage
	ldi YL,LOW(DISPLAY_IF)
	ldi YH,HIGH(DISPLAY_IF)
	ldi r16,USART_CMODE_MSPI_gc
	std Y+USART_CTRLC_offset,r16
	clr r16										;full speed
	std Y+USART_BAUDCTRLA_offset,r16
	std Y+USART_BAUDCTRLB_offset,r16
	ldi r16,USART_TXEN_bm
	std Y+USART_CTRLB_offset,r16

	.if spiram_rd==1
;initialize interface towards SPI RAM
	.if RAM_IF_MODE==1
;USART in master SPI mode
	ldi YL,LOW(RAM_IF)
	ldi YH,HIGH(RAM_IF)
	ldi r16,USART_CMODE_MSPI_gc
	std Y+USART_CTRLC_offset,r16
	.if clock>48
	ldi r16,1									;23LC1024 doesn't support clock frequency >20MHz
	.else
	clr r16
	.endif
	std Y+USART_BAUDCTRLA_offset,r16
	clr r16
	std Y+USART_BAUDCTRLB_offset,r16
	ldi r16,USART_TXEN_bm|USART_RXEN_bm
	std Y+USART_CTRLB_offset,r16
	.else
;native SPI
	.if clock>48
	ldi r16,SPI_ENABLE_bm|SPI_MASTER_bm|SPI_PRESCALER_DIV4_gc	;23LC1024 doesn't support clock frequency >20MHz
	.else
	ldi r16,SPI_CLK2X_bm|SPI_ENABLE_bm|SPI_MASTER_bm|SPI_PRESCALER_DIV4_gc
	.endif
	sts (RAM_IF+SPI_CTRL_offset),r16
	.endif

	lds r16,(MODE_PIN_PORT)
	sbrs r16,MODE_PIN
	jmp load_ram								;jump to loader for external RAM
	.endif

restart: ldi r16,-1
	sts (DISPLAY_IF+USART_DATA_offset),r16
	rcall lcd_spi_wt

;initialize LCD
	ldi ZL,LOW(disp_init_str*2)
	ldi ZH,HIGH(disp_init_str*2)
disp_init_lp1: in r16,DISPLAYPORT
	andi r16,~(1<<DISPLAY_SS|1<<DISPLAY_C_D)
	out DISPLAYPORT,r16							;assert /SS and set LCD command mode
disp_init_lp: lpm r16,Z+
	tst r16
	breq disp_initx
	rcall lcd_wr_spi
	lpm r16,Z+
	tst r16
	breq disp_init_lp
	cpi r16,-1
	brne disp_init1
	ldi r16,10
	rcall wait
	rjmp disp_init_lp
disp_init1: mov r17,r16
	rcall lcd_spi_wt
	sbi DISPLAYPORT,DISPLAY_C_D					;set LCD data mode
disp_init_lp2: lpm r16,Z+
	rcall lcd_wr_spi
	dec r17
	brne disp_init_lp2
	rcall lcd_spi_wt
	rjmp disp_init_lp1
disp_initx:
	in r16,DISPLAYPORT
	ori r16,1<<DISPLAY_SS|1<<DISPLAY_C_D		;release /SS and /C for LCD
	out DISPLAYPORT,r16

	.if spiram_rd==1
	clr r18										;set address register in serial RAM to 0
	clr r19
	clr r20
	sts (fpos),r18
	rcall set_rampos
	rcall ram_wr_rd_spi							;read first two bytes from serial RAM
	mov r17,r16
	rcall ram_wr_rd_spi
	sbi RAMPORT,RAMPORT_SS						;release /SS for external RAM
	cpi r17,$55									;check magic values $55/$aa if serial RAM is loaded with data
	brne stop									;abort if no image data loaded into serial RAM
	cpi r16,$aa
	brne stop
	.endif
	.if flash_rd==1
	clr r16
	sts (fpos),r16
	.endif
loop:
	.if usart_rd==1
	clr r16										;reset serial buffer pointers
	sts (rd_sptr),r16
	sts (wr_sptr),r16
	cbi RTS										;enable RTS
	.endif
	.if spiram_rd==1
	lds r21,(fpos)								;get current position in the files list
loop1: ldi r16,4
	mul r16,r21
	inc r21
	sts (fpos),r21
	movw r19:r18,r1:r0
	subi r18,-2									;add offset of two (reserved for the magic values)
	sbci r19,-1
	clr r20
	rcall set_rampos
	rcall ram_wr_rd_spi							;get four bytes from external RAM
	mov r18,r16									;low address byte
	rcall ram_wr_rd_spi
	mov r19,r16									;middle address byte
	rcall ram_wr_rd_spi
	mov r20,r16									;high address byte
	rcall ram_wr_rd_spi
	mov r22,r16									;delay value
	sbi RAMPORT,RAMPORT_SS						;release /SS for external RAM
	or r16,r18									;check if address==0
	or r16,r19
	or r16,r20
	brne x1										;no, proceed with image decoding
	clr r21										;restart with first image
	rjmp loop1	
x1: push r22
	rcall set_rampos							;set image file address in external RAM
	.if RAM_IF_MODE==1
	sts (RAM_IF+USART_DATA_offset),r16			;prepare delivery of first data byte
	.else
	sts (RAM_IF+SPI_DATA_offset),r16
	.endif
	.endif

	.if flash_rd==1
	lds r21,(fpos)								;get current position in the files list
loop1: ldi r16,4
	mul r16,r21
	inc r21
	sts (fpos),r21
	ldi ZL,LOW(files*2)							;add start address of files list
	add ZL,r0
	ldi ZH,HIGH(files*2)
	adc ZH,r1
	.if big_flash==1
	ldi r17,0
	ldi r16,PAGE(files*2)
	adc r16,r17
	out CPU_RAMPZ,r16
	elpm r18,Z+									;get address of picture file
	elpm r19,Z+
	elpm r20,Z+
	elpm r22,Z									;get delay
	mov r16,r18
	or r16,r19
	or r16,r20									;check if address==0
	.else
	lpm r18,Z+									;get address of picture file
	lpm r19,Z+
	adiw ZH:ZL,1
	lpm r22,Z									;get delay
	mov r16,r18
	or r16,r19									;check if address==0
	.endif
	brne x2
	clr r21										;if address==0 restart with first picture file
	rjmp loop1
x2:	sts (rd_ptr),r18
	sts (rd_ptr+1),r19
	.if big_flash==1
	sts (rd_ptr+2),r20
	.endif
	push r22
	.endif

;decode and show image
	rcall dec_jpg

	.if usart_rd==1
;wait 250 milliseconds for trailing bytes over serial interface
	ldi r16,250
	rcall wait
	rjmp loop
	.endif
	.if spiram_rd==1
	sbi RAMPORT,RAMPORT_SS						;release /SS of external RAM
	.if RAM_IF_MODE==1
	ldi r16,-1
	sts (RAM_IF+USART_STATUS_offset),r16		;clear status of interface to external RAM
	lds r16,(RAM_IF+USART_DATA_offset)			;empty receive buffer
	.else
	lds r16,(RAM_IF+SPI_STATUS_offset)			;clear status of interface to external RAM
	lds r16,(RAM_IF+SPI_DATA_offset)			;empty receive buffer
	.endif
	.endif
	.if spiram_rd==1 || flash_rd==1
	pop r22
	mul r22,r22									;wait (r22)^2 milliseconds
	breq loop									;don't wait if r22==0
	movw r19:r18,r1:r0
wt_lp: mov r16,r18
	subi r18,100
	sbci r19,0
	brcs wtx
	ldi r16,100
	rcall wait
	rjmp wt_lp
wtx: rcall wait
	rjmp loop
	.endif

stop: rjmp stop

;decrement bit counter and feed next byte of huffman coded image data into the JPEG decoder
;capable to skip restart markers, but not other marker segments
;input:	r16			bit stream buffer
;		r17			count of remaining bits
;output: r16		adapted bit stream buffer
;		r17			adapted count of remaining bits
	.macro get_nxt_byte
	dec r17										;decrement bit stream counter
	brne get_nxt_bytex							;skip if still bits left
	ldi r17,8
get_nxt_byte_rld: rd_byte						;get next data byte
	cpi r16,$ff									;possible start of marker?
	brne get_nxt_bytex							;no; deliver data byte
	rd_byte										;get next data byte
	tst r16										;2nd byte==0 for byte stuffing?
	brne get_nxt_byte_rld						;no, assume restart marker and get next byte 
	ldi r16,$ff									;deliver original value
get_nxt_bytex:
	.endmacro

;general read macro, delivers single byte input to JPEG decoder
;output: r16		byte read from source
	.macro rd_byte
	.if flash_rd==1
	rcall rd_byte_flash
	.endif
	.if usart_rd==1
	rcall rd_byte_sbuf
	.endif
	.if spiram_rd==1
	.if RAM_IF_MODE==1
;USART in master SPI mode
spimem_rd: lds r16,(RAM_IF+USART_STATUS_offset)
	andi r16,USART_TXCIF_bm
	breq spimem_rd
	sts (RAM_IF+USART_STATUS_offset),r16
	lds r16,(RAM_IF+USART_DATA_offset)
	sts (RAM_IF+USART_DATA_offset),r16			;trigger reading of next byte
	.else
;native SPI
spimem_rd: lds r16,(RAM_IF+SPI_STATUS_offset)
	andi r16,SPI_IF_bm
	breq spimem_rd
	lds r16,(RAM_IF+SPI_DATA_offset)
	sts (RAM_IF+SPI_DATA_offset),r16			;trigger reading of next byte
	.endif
	.endif
	.endmacro

;JPEG decoder, decodes image and shows it on LCD
;first part, preparation of the tables and basic values
dec_jpg: ldi r16,LOW(ht)
	ldi r17,HIGH(ht)
	movw r15:r14,r17:r16						;r15:r14 hold start address of Huffman tree to be built
file_scan_lp: rd_byte
	cpi r16,$ff									;check if 1st byte of marker
	breq file_scan
file_error: ret

file_scan: rd_byte
	cpi r16,$d8									;SOI
	breq file_scan_lp
	mov r17,r16
	rd_byte										;read segment length parameter
	mov r25,r16
	rd_byte
	mov r24,r16
	sbiw r25:r24,2
	cpi r17,$db									;DQT
	brne file_scan1
dqt_lp: rd_byte
	ldi r17,$f0
	and r17,r16
	brne file_error								;only 8 bit precision supported
	andi r16,15
	ldi r17,64
	mul r16,r17
	ldi YL,LOW(q_tbl)
	ldi YH,HIGH(q_tbl)
	add YL,r0
	adc YH,r1
	ldi r18,64									;read 64 quantization values
dqt_cp: rd_byte
	st Y+,r16
	dec r18
	brne dqt_cp
	subi r24,65
	brne dqt_lp
	rjmp file_scan_lp

file_scan1: cpi r17,$c4							;DHT
	brne file_scan2
	movw YH:YL,r15:r14
	clr r6
dht_scan: rcall build_ht						;build one huffman deconding tree
	mov r16,r24
	or r16,r25
	brne dht_scan								;continue if more data left
	movw r15:r14,YH:YL
	rjmp file_scan_lp
file_error_jp: rjmp file_error

file_scan3: cpi r17,$da							;SOS
	brne file_skip
	rjmp decode
file_skip: rd_byte								;skip any other segment
	sbiw r25:r24,1
	brne file_skip
	rjmp file_scan_lp

file_scan2: cpi r17,$c0							;SOF0
	brne file_scan3
	rd_byte
	cpi r16,8									;only 8 bit data precision supported
file_error2: brne file_error_jp
	ldi r17,4
skip_hw: rd_byte								;skip height and width
	dec r17
	brne skip_hw
	rd_byte
	cpi r16,3									;only 3 components supported
	brne file_error2
	subi r24,6
	clr r19
sof0_lp: rd_byte
	mov r17,r16									;component ID
	rd_byte
	mov r18,r16									;sampling factors
	rd_byte
	mov r0,r16									;quantization table number
	cpi r17,1
	brne sof0_3									;skip if component ID not Y
	ldi r17,15
	and r17,r18
	cpi r17,1									;check if vertical sampling factor !=1
	breq sof0_1
	ori r19,1
sof0_1: andi r18,$f0
	cpi r18,$10									;check if horizontal sampling factor !=1
	breq sof0_3
	ori r19,2
sof0_3: subi r24,3
	brne sof0_lp
	sts (sfact),r19								;(sfact) bit 0 set: vertical sampling factor=2
												;(sfact) bit 1 set: horizontal sampling factor=2
	rjmp file_scan_lp

;decodes huffman coded image data and shows it on LCD
decode:
	.if lcd_dma==1
;global settings for DMA needed only once to be defined
	clr r0
	ldi YL,LOW(DMA_BASE)
	ldi YH,HIGH(DMA_BASE)
	ldi r16,DMA_CH_SRCDIR0_bm						;source address will be incremented during active DMA
	std Y+DMA_CH0_offset+DMA_CH_ADDRCTRL_offset,r16
	ldi r16,DISPLAY_DMA_TRIG						;trigger source USART data register empty
	std Y+DMA_CH0_offset+DMA_CH_TRIGSRC_offset,r16
	lds r17,(sfact)
	ldi r16,3*8										;8 RGB pixels with 3 bytes each
	tst r17
	breq dma_length1
	add r16,r16										;double pixel count if one of the sampling factors >1
	cpi r17,3
	brne dma_length1
	add r16,r16										;double pixel count if both of the sampling factors >1
dma_length1:
	std Y+DMA_CH0_offset+DMA_CH_TRFCNT_offset,r16	;block transfer count, number of bytes to be transferred
	std Y+DMA_CH0_offset+DMA_CH_TRFCNT_offset+1,r0
	ldi r16,LOW(DISPLAY_IF+USART_DATA_offset)		;address of LCD SPI port
	std Y+DMA_CH0_offset+DMA_CH_DESTADDR0_offset,r16
	ldi r16,HIGH(DISPLAY_IF+USART_DATA_offset)
	std Y+DMA_CH0_offset+DMA_CH_DESTADDR1_offset,r16
	std Y+DMA_CH0_offset+DMA_CH_DESTADDR2_offset,r0
	ldi r16,DMA_ENABLE_bm							;globally enable DMA (actual transfer not started here)
	std Y+DMA_CTRL_offset,r16
	.endif

;skip component information preceding image data assuming components are assigned the usual way
skip_sos_hd: rd_byte
	sbiw r25:r24,1
	brne skip_sos_hd
;DC value per component is coded as difference to previous block
;initialize with zero for Y, Cb and Cr
	ldi YL,LOW(prv_dc)
	ldi YH,HIGH(prv_dc)
	ldi r16,3*2
clr_prv_dc: st Y+,r2
	dec r16
	brne clr_prv_dc
	ldi r17,1

	clr r25										;line block counter, one block = 8 pixel lines
scan_loop: push r25
	ldi r19,8
	mul r25,r19
	movw XH:XL,r1:r0							;r1:r0 upper row on LCD
	adiw X,7									;XH:XL lower row on LCD
	lds r19,(sfact)
	sbrc r19,0
	adiw X,8									;double area if vertical sampling factor >1
	ldi r21,$2b									;LCD command "Page Address Set"
	rcall lcd_set_edge							;Y 
	clr r24										;column block counter, one block = 8 pixel columns
scan_loop1: push r24
	ldi r19,8
	mul r24,r19
	movw XH:XL,r1:r0							;r1:r0 left column on LCD
	adiw X,7									;XH:XL right column on LCD
	lds r19,(sfact)
	sbrc r19,1
	adiw X,8									;double area if horizontal sampling factor >1
	ldi r21,$2a									;LCD command "Column Address Set"
	rcall lcd_set_edge							;X

;read data for one MCU
;an MCU consists of one Cb and Cr block and one, two or four Y blocks, depending on the sampling factors
	clr r2
	clr r19
	clr r21
	ldi r22,LOW(prv_dc)
	ldi r23,HIGH(prv_dc)
	ldi YL,LOW(blk_y)
	ldi YH,HIGH(blk_y)
	rcall get_blk								;read 1st Y block

	lds r18,(sfact)
	tst r18
	breq scan_yx								;continue with Cb block if no chroma subsampling

	clr r19
	clr r21
	ldi YL,LOW(blk_y+128)
	ldi YH,HIGH(blk_y+128)
	rcall get_blk								;read 2nd Y block 

	lds r18,(sfact)
	cpi r18,3
	brne scan_yx								;continue with Cb block except both sampling factors are >1

	clr r19
	clr r21
	ldi YL,LOW(blk_y+256)
	ldi YH,HIGH(blk_y+256)
	rcall get_blk								;read 3rd Y block

	clr r19
	clr r21
	ldi YL,LOW(blk_y+384)
	ldi YH,HIGH(blk_y+384)
	rcall get_blk								;read 4th Y block

scan_yx:

	ldi r19,4
	ldi r21,64
	ldi r22,LOW(prv_dc+2)
	ldi r23,HIGH(prv_dc+2)
	ldi YL,LOW(blk_cb)
	ldi YH,HIGH(blk_cb)
	rcall get_blk								;read one Cb block

	ldi r19,4
	ldi r21,64
	ldi r22,LOW(prv_dc+4)
	ldi r23,HIGH(prv_dc+4)
	ldi YL,LOW(blk_cr)
	ldi YH,HIGH(blk_cr)
	rcall get_blk								;read one Cr block

	push r16
	push r17
	push ZL
	push ZH

	ldi YL,LOW(blk_y)
	ldi YH,HIGH(blk_y)
	rcall idct									;transform 1st Y block

	ldi YL,LOW(blk_cb)
	ldi YH,HIGH(blk_cb)
	rcall idct									;transform Cb block

	ldi YL,LOW(blk_cr)
	ldi YH,HIGH(blk_cr)
	rcall idct									;transform Cr block

	lds r16,(sfact)
	tst r16
	breq scan_conv								;no more transformations if no chroma subsampling
	
	ldi YL,LOW(blk_y+128)
	ldi YH,HIGH(blk_y+128)
	rcall idct									;transform 2nd Y block

	lds r16,(sfact)
	cpi r16,3
	brne scan_conv								;no more transformations except both sampling factors are >1

	ldi YL,LOW(blk_y+256)
	ldi YH,HIGH(blk_y+256)
	rcall idct									;transform 3rd Y block

	ldi YL,LOW(blk_y+384)
	ldi YH,HIGH(blk_y+384)
	rcall idct									;transform 4th Y block

scan_conv:

	in r16,DISPLAYPORT
	andi r16,~(1<<DISPLAY_SS|1<<DISPLAY_C_D)
	out DISPLAYPORT,r16							;assert /SS and set LCD command mode
	ldi r16,$2c									;LCD command "Memory Write"
	rcall lcd_wr_spi
	rcall lcd_spi_wt							;wait until command has been sent to LCD
	sbi DISPLAYPORT,DISPLAY_C_D					;set LCD data mode

	rcall rgb_conv								;convert MCU to RGB data and send it to LCD

	.if lcd_dma==1
dma_wait1: lds r16,(DMA_STATUS)
	sbrc r16,DMA_CH0BUSY_bp
	rjmp dma_wait1								;wait until last DMA block transfer has been finished
	.endif

	rcall lcd_spi_wt							;wait until last data byte has been sent to LCD
	sbi DISPLAYPORT,DISPLAY_SS					;release /SS for LCD

	pop ZH
	pop ZL
	pop r17
	pop r16
	pop r24

	lds r18,(sfact)
	inc r24
	sbrc r18,1									;double step if horizontal sampling factor >1
	inc r24
	cpi r24,x_/8
	breq scan1
	rjmp scan_loop1
scan1: pop r25
	inc r25
	sbrc r18,0									;double step if vertical sampling factor >1
	inc r25
	cpi r25,y_/8
	breq scan_x
	rjmp scan_loop
scan_x: ret

;trailer for 4:2:2 vertical chroma subsampling
rgb_cnv_422v:
	adiw X,7*3									;set address to next line in output buffer
	ldd r2,Y+16									;get 2nd Y value (original X position on line below)
	rcall rgb_triple							;convert 2nd pixel to RGB
	sbiw X,8*3									;set address to original line in output buffer
	subi r24,-2									;point to next Y entry
	dec r18
	brne rgb_conv_lp
	subi r24,-$10								;continue with next line in "upper" Y block
	rcall lcd_send_buf							;send current line of RGB pixels to LCD
	dec r19
	brne rgb_conv_lp1
	ret

;trailer for 4:2:2 horizontal chroma subsampling
rgb_cnv_422h:
	ldd r2,Y+2									;get 2nd Y value (next on same line)
	rcall rgb_triple							;convert 2nd pixel to RGB
	subi r24,-4									;point to next Y entry
	cpi r18,5									;check if first four columns (two pixels per column) were finished
	brne rgb_cnv_422h_1
	subi r24,-$70								;advance to "right" Y block
rgb_cnv_422h_1: dec r18
	brne rgb_conv_lp
	subi r24,$80								;revert to "left" Y block
	rcall lcd_send_buf							;send current line of RGB pixels to LCD
	dec r19
	brne rgb_conv_lp1
	ret

;trailer for 4:4:4 no chroma subsampling
rgb_cnv_444:
	subi r24,-2									;point to next Y entry
	dec r18
	brne rgb_conv_lp
	rcall lcd_send_buf							;send current line of RGB pixels to LCD
	dec r19
	brne rgb_conv_lp1
	ret

;convert all pixels of an MCU to RGB values and send them to LCD
rgb_conv:
	clr r3
	ldi r22,LOW(tmp_buf)
	ldi r23,HIGH(tmp_buf)
	ldi ZL,LOW(rgb_cjmp_tbl*2)
	ldi ZH,HIGH(rgb_cjmp_tbl*2)
	lds r16,(sfact)
	.if lcd_dma==0
	mov r17,r16
	.endif
	add r16,r16
	add ZL,r16
	adc ZH,r3
	lpm r4,Z+									;r5:r4 holds address of trailer depending on the chroma subsampling mode
	lpm r5,Z
	.if lcd_dma==0
	ldi r16,3*8									;8 RGB pixels with 3 bytes each
	tst r17
	breq buf_length1
	add r16,r16									;double pixel count if one of the sampling factors >1
	cpi r17,3
	brne buf_length1
	add r16,r16									;double pixel count if both of the sampling factors >1
buf_length1: mov r9,r16
	.endif
	ldi r24,LOW(blk_y)
	ldi r25,HIGH(blk_y)
	ldi r20,LOW(blk_cb)
	ldi r21,HIGH(blk_cb)
	ldi r19,8									;line count
rgb_conv_lp1: ldi r18,8							;column count
	movw XH:XL,r23:r22
rgb_conv_lp: 
	movw YH:YL,r21:r20
	subi r20,-2
	ld ZL,Y										;get Cb value
	ldi ZH,HIGH(rgb_cb_tbl0_344*2)
	lpm r0,Z									;convert Cb factor for green value calculation 
	clr r1
	sbrc r0,7
	com r1										;sign extension to 16 bit value
	ldi ZH,HIGH(rgb_cb_tbl1_772)
	add ZL,ZL
	adc ZH,ZH
	lpm r10,Z+									;convert Cb factor for blue value calculation
	lpm r11,Z
	subi YL,LOW(-(blk_cr-blk_cb))
	ld ZL,Y										;get Cr value
	ldi ZH,HIGH(rgb_cr_tbl0_714*2)
	lpm r12,Z									;convert Cr factor for green value calculation
	clr r13
	sbrc r12,7
	com r13										;sign extension to 16 bit value
	add r12,r0									;summed factor for green value calculation
	adc r13,r1
	ldi ZH,HIGH(rgb_cr_tbl1_402)
	add ZL,ZL
	adc ZH,ZH
	lpm r14,Z+									;convert Cr factor for red value calculation
	lpm r15,Z
	movw YH:YL,r25:r24
	ld r2,Y										;get 1st Y value
	call rgb_triple								;convert 1st pixel to RGB
	movw ZH:ZL,r5:r4
	ijmp										;jump to trailer to optionally handle further pixels depending on chroma subsampling mode

;address table of trailer routines
rgb_cjmp_tbl: .dw rgb_cnv_444,rgb_cnv_422v,rgb_cnv_422h,rgb_cnv_420

;trailer for 4:2:0 horizontal/vertical chroma subsampling
rgb_cnv_420: 
	ldd r2,Y+2									;get 2nd Y value (next on same line)
	rcall rgb_triple							;convert 2nd pixel to RGB
	adiw X,14*3									;set address to next line in output buffer
	ldd r2,Y+16									;get 3rd Y value (original X position on line below)
	rcall rgb_triple							;convert 3rd pixel to RGB
	ldd r2,Y+18									;get 4th Y value (next on line below)
	rcall rgb_triple							;convert 4th pixel to RGB
	sbiw X,16*3									;set address to original line in output buffer
	subi r24,-4									;point to next Y entry
	cpi r18,5									;check if first four columns (two pixels per column) were finished
	brne rgb_cnv_420_1
	subi r24,-$70								;advance to "right" Y block
rgb_cnv_420_1: dec r18
	brne rgb_conv_lp
	subi r24,$70								;revert to "left" Y block
	cpi r24,$80									;"upper" Y blocks finished?
	brne rgb_cnv_420_2
	clr r24										;continue with "lower" Y blocks
	inc r25
rgb_cnv_420_2:
	rcall lcd_send_buf							;send current lines of RGB pixels to LCD
	dec r19
	brne rgb_conv_lp1
	ret

	.macro rgb_bclip
	breq rgb_bc1
	ldi r16,0
	brmi rgb_bc1
	com r16
rgb_bc1:
	.endmacro

;convert single pixel to RGB
rgb_triple:
	movw r17:r16,r3:r2
	add r16,r14
	adc r17,r15
	rgb_bclip
	st X+,r16
	movw r17:r16,r3:r2
	add r16,r12
	adc r17,r13
	rgb_bclip
	st X+,r16
	movw r17:r16,r3:r2
	add r16,r10
	adc r17,r11
	rgb_bclip
	st X+,r16
	ret

;prepare one block (8 x 8 values) to be fed into IDCT calculation
;including decompression, dequantization, scaling and re-ordering
;
;input:	r19			pointer offset to corresponding Huffman tree pair
;		r21			offset to corresponding quantization table		
;		r23:r22		pointer to stored DC value of previous block
;		YH:YL		pointer to component block storage
get_blk:
	subi r21,LOW(-q_tbl)
	mov r4,r21
	clr r21
	sbci r21,HIGH(-q_tbl)
	mov r5,r21									;r5:r4 holds address of quantization table
	push YL
	push YH
	ldi r24,LOW(pow2tab*2-2)
	ldi r25,HIGH(pow2tab*2-2)
	movw r11:r10,r25:r24						;r11:r10 points to a table holding power of 2 values for decoding AC/DC values

	ldi YL,LOW(ht_ptr)
	ldi YH,HIGH(ht_ptr)
	add YL,r19
	adc YH,r2
	ld r24,Y+
	ld r25,Y+									;r25:r24 holds the address of the appropriate Huffman decoding tree for DC
	push YL
	push YH

;clear temporary output buffer at once so that empty ranges can be skipped during decompression instead of clearing them individually
	ldi YL,LOW(tmp_buf)
	ldi YH,HIGH(tmp_buf)
	movw XH:XL,YH:YL							;XH:XL holds pointer to temporary output buffer
	ldi r20,128/32
clr_blk_lp: 
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	dec r20
	brne clr_blk_lp

	rcall get_node								;find corresponding node entry for DC value
	rcall get_value								;get remaining bits and convert to signed value
	movw YH:YL,r23:r22
	ld r18,Y									;get previous DC value
	ldd r19,Y+1
	add r20,r18									;actual DC value is difference between previous value and the one just read
	adc r21,r19
	st Y,r20									;remember for next time
	std Y+1,r21
	movw YH:YL,r5:r4
	ld r19,Y+									;get quantization value from table
	mulsu r21,r19								;apply dequantization
	mov r3,r0
	mul r20,r19
	add r1,r3
	movw r5:r4,YH:YL
	st X+,r0
	st X+,r1

	ldi r18,63
	mov r12,r18
	ldi r24,LOW(arai_scale_zz*2)
	ldi r25,HIGH(arai_scale_zz*2)
	movw r7:r6,r25:r24							;r7:r6 points to current entry in table of Arai scaling factors
	pop YH
	pop YL
	ld r24,Y+
	ld r25,Y									;r25:r24 now holds the address of the appropriate Huffman decoding tree for AC
get_ac_lp:
	rcall get_node								;find corresponding node entry for AC value
	tst r19										;EOB node found?
	breq get_acx								;yes, decompression for this block is finished
	rcall get_value								;get remaining bits and convert to signed value, skipping range of zero coded values
	movw YH:YL,r5:r4
	ld r19,Y+									;get quantization value from table
	mulsu r21,r19								;apply dequantization
	mov r3,r0
	mul r20,r19
	add r1,r3
	movw r21:r20,r1:r0
	movw r5:r4,YH:YL

;	rjmp skip_arai_scl							;skip Arai scaling in case of troubleshooting to compare calculated values with JPEGSnoop's output

	movw ZH:ZL,r7:r6
	lpm r18,Z+									;get Arai scaling factor
	lpm r19,Z+
	movw r7:r6,ZH:ZL

;apply prescaling according to Arai's IDCT algorithm
;	alternatively the quantization tables could be prepared accordingly once before,
;	saving the need of an extra multiplication here; it might slightly increase decoding speed
;
;	but to retain accuracy with regards to memory (RAM) consumption it's done this way
	fmulsu r21,r19
	movw r9:r8,r1:r0
	fmul r20,r18
	adc r8,r2
	mov r3,r1
	fmulsu r21,r18
	sbc r9,r2
	add r3,r0
	adc r8,r1
	adc r9,r2
	fmul r20,r19
	adc r9,r2
	add r3,r0
	adc r8,r1
	adc r9,r2
	movw r21:r20,r9:r8

skip_arai_scl:
	st X+,r20
	st X+,r21
	dec r12
	brne get_ac_lp
get_acx:
	pop XH
	pop XL										;XH:XL points to component block storage

;copy values from temporary buffer to component block, reverting the zigzag order
	ldi YL,LOW(tmp_buf)
	ldi YH,HIGH(tmp_buf)
	ldi ZL,LOW(tmp_buf+64)
	ldi ZH,HIGH(tmp_buf+64)
;0
	ld r18,Y
	ldd r19,Y+1
	st X+,r18
	st X+,r19
;1
	ldd r18,Y+2
	ldd r19,Y+3
	st X+,r18
	st X+,r19
;5
	ldd r18,Y+10
	ldd r19,Y+11
	st X+,r18
	st X+,r19
;6
	ldd r18,Y+12
	ldd r19,Y+13
	st X+,r18
	st X+,r19
;14
	ldd r18,Y+28
	ldd r19,Y+29
	st X+,r18
	st X+,r19
;15
	ldd r18,Y+30
	ldd r19,Y+31
	st X+,r18
	st X+,r19
;27
	ldd r18,Y+54
	ldd r19,Y+55
	st X+,r18
	st X+,r19
;28
	ldd r18,Y+56
	ldd r19,Y+57
	st X+,r18
	st X+,r19
;2
	ldd r18,Y+4
	ldd r19,Y+5
	st X+,r18
	st X+,r19
;4
	ldd r18,Y+8
	ldd r19,Y+9
	st X+,r18
	st X+,r19
;7
	ldd r18,Y+14
	ldd r19,Y+15
	st X+,r18
	st X+,r19
;13
	ldd r18,Y+26
	ldd r19,Y+27
	st X+,r18
	st X+,r19
;16
	ldd r18,Y+32
	ldd r19,Y+33
	st X+,r18
	st X+,r19
;26
	ldd r18,Y+52
	ldd r19,Y+53
	st X+,r18
	st X+,r19
;29
	ldd r18,Y+58
	ldd r19,Y+59
	st X+,r18
	st X+,r19
;42
	ldd r18,Z+20
	ldd r19,Z+21
	st X+,r18
	st X+,r19
;3
	ldd r18,Y+6
	ldd r19,Y+7
	st X+,r18
	st X+,r19
;8
	ldd r18,Y+16
	ldd r19,Y+17
	st X+,r18
	st X+,r19
;12
	ldd r18,Y+24
	ldd r19,Y+25
	st X+,r18
	st X+,r19
;17
	ldd r18,Y+34
	ldd r19,Y+35
	st X+,r18
	st X+,r19
;25
	ldd r18,Y+50
	ldd r19,Y+51
	st X+,r18
	st X+,r19
;30
	ldd r18,Y+60
	ldd r19,Y+61
	st X+,r18
	st X+,r19
;41
	ldd r18,Z+18
	ldd r19,Z+19
	st X+,r18
	st X+,r19
;43
	ldd r18,Z+22
	ldd r19,Z+23
	st X+,r18
	st X+,r19
;9
	ldd r18,Y+18
	ldd r19,Y+19
	st X+,r18
	st X+,r19
;11
	ldd r18,Y+22
	ldd r19,Y+23
	st X+,r18
	st X+,r19
;18
	ldd r18,Y+36
	ldd r19,Y+37
	st X+,r18
	st X+,r19
;24
	ldd r18,Y+48
	ldd r19,Y+49
	st X+,r18
	st X+,r19
;31
	ldd r18,Y+62
	ldd r19,Y+63
	st X+,r18
	st X+,r19
;40
	ldd r18,Z+16
	ldd r19,Z+17
	st X+,r18
	st X+,r19
;44
	ldd r18,Z+24
	ldd r19,Z+25
	st X+,r18
	st X+,r19
;53
	ldd r18,Z+42
	ldd r19,Z+43
	st X+,r18
	st X+,r19
;10
	ldd r18,Y+20
	ldd r19,Y+21
	st X+,r18
	st X+,r19
;19
	ldd r18,Y+38
	ldd r19,Y+39
	st X+,r18
	st X+,r19
;23
	ldd r18,Y+46
	ldd r19,Y+47
	st X+,r18
	st X+,r19
;32
	ld r18,Z
	ldd r19,Z+1
	st X+,r18
	st X+,r19
;39
	ldd r18,Z+14
	ldd r19,Z+15
	st X+,r18
	st X+,r19
;45
	ldd r18,Z+26
	ldd r19,Z+27
	st X+,r18
	st X+,r19
;52
	ldd r18,Z+40
	ldd r19,Z+41
	st X+,r18
	st X+,r19
;54
	ldd r18,Z+44
	ldd r19,Z+45
	st X+,r18
	st X+,r19
;20
	ldd r18,Y+40
	ldd r19,Y+41
	st X+,r18
	st X+,r19
;22
	ldd r18,Y+44
	ldd r19,Y+45
	st X+,r18
	st X+,r19
;33
	ldd r18,Z+2
	ldd r19,Z+3
	st X+,r18
	st X+,r19
;38
	ldd r18,Z+12
	ldd r19,Z+13
	st X+,r18
	st X+,r19
;46
	ldd r18,Z+28
	ldd r19,Z+29
	st X+,r18
	st X+,r19
;51
	ldd r18,Z+38
	ldd r19,Z+39
	st X+,r18
	st X+,r19
;55
	ldd r18,Z+46
	ldd r19,Z+47
	st X+,r18
	st X+,r19
;60
	ldd r18,Z+56
	ldd r19,Z+57
	st X+,r18
	st X+,r19
;21
	ldd r18,Y+42
	ldd r19,Y+43
	st X+,r18
	st X+,r19
;34
	ldd r18,Z+4
	ldd r19,Z+5
	st X+,r18
	st X+,r19
;37
	ldd r18,Z+10
	ldd r19,Z+11
	st X+,r18
	st X+,r19
;47
	ldd r18,Z+30
	ldd r19,Z+31
	st X+,r18
	st X+,r19
;50
	ldd r18,Z+36
	ldd r19,Z+37
	st X+,r18
	st X+,r19
;56
	ldd r18,Z+48
	ldd r19,Z+49
	st X+,r18
	st X+,r19
;59
	ldd r18,Z+54
	ldd r19,Z+55
	st X+,r18
	st X+,r19
;61
	ldd r18,Z+58
	ldd r19,Z+59
	st X+,r18
	st X+,r19
;35
	ldd r18,Z+6
	ldd r19,Z+7
	st X+,r18
	st X+,r19
;36
	ldd r18,Z+8
	ldd r19,Z+9
	st X+,r18
	st X+,r19
;48
	ldd r18,Z+32
	ldd r19,Z+33
	st X+,r18
	st X+,r19
;49
	ldd r18,Z+34
	ldd r19,Z+35
	st X+,r18
	st X+,r19	
;57
	ldd r18,Z+50
	ldd r19,Z+51
	st X+,r18
	st X+,r19
;58
	ldd r18,Z+52
	ldd r19,Z+53
	st X+,r18
	st X+,r19
;62
	ldd r18,Z+60
	ldd r19,Z+61
	st X+,r18
	st X+,r19
;63
	ldd r18,Z+62
	ldd r19,Z+63
	st X+,r18
	st X+,r19
	ret

;deliver actual value based on previously determined code entry from Huffman decoding tree
;skip range of zero coded values
;input:	r5:r4		pointer to current entry in quantization table
;		r7:r6		pointer to current entry in Arai scaling factor table
;		r11:r10		pointer to power of 2 table
;		r12			number of remaining values in current block
;		r16			bit stream buffer
;		r17			number of valid bits in bit stream buffer
;		r19			code entry from Huffman decoding tree
;		XH:XL		pointer to current position in temporary output buffer
;output: r5:r4		adapted pointer to current entry in quantization table
;		r7:r6		adapted pointer to current entry in Arai scaling factor table
;		r12			adapted number of remaining values in current block
;		r16			adapted bit stream buffer
;		r17			adapted number of valid bits in bit stream buffer
;		r21:r20		signed AC/DC value
;		XH:XL		adapted pointer to current position in temporary output buffer
get_value: clr r20
	clr r21
	ldi r18,$f0
	and r18,r19
	breq get_value1								;no range of zero values to skip
	swap r18									;r18 holds number of zero values to skip
	sub r12,r18
	add r4,r18
	adc r5,r2
	add r18,r18
	add XL,r18
	adc XH,r2
	add r6,r18
	adc r7,r2
get_value1: andi r19,15
	breq get_valuex								;skip if bit count of output value is zero
	mov r3,r19
get_val_lp: get_nxt_byte
	lsl r16
	rol r20
	rol r21
	dec r19
	brne get_val_lp
	movw ZH:ZL,r11:r10
	add r3,r3
	add ZL,r3
	adc ZH,r2
	lpm r18,Z+									;get power of 2 value for bit count
	lpm r19,Z
	movw r1:r0,r19:r18
	and r0,r20
	and r1,r21
	or r0,r1
	brne get_valuex								;skip if positive value
	add r18,r18									;invert for negative value
	adc r19,r19
	subi r18,1
	sbci r19,0
	com r19
	neg r18
	sbci r19,-1
	add r20,r18
	adc r21,r19
get_valuex: ret

;read bits from input stream, find corresponding node in Huffman decoding tree
;input:	r16			bit stream buffer
;		r17			number of remaining bits in bit stream buffer
;		r25:r24		address of Huffman decoding tree
;output: r16		adapted bit stream buffer
;		r17			adapted number of remaining bits in bit stream buffer
;		r19			code entry from Huffman decoding tree
get_node: movw YH:YL,r25:r24
	ld r18,Y+									;get distance to next tree level
get_node_lp: get_nxt_byte
	lsl r16
	brcc get_node2								;skip if current bit was 0
	subi r18,-2									;add two if current bit was set
get_node2: add YL,r18							;add distance to pointer
	adc YH,r2
	ld r18,Y+									;get current distance
	tst r18
	brne get_node_lp							;repeat if distance >0
	ld r19,Y									;get code value
	ret

;construct Huffman decoding tree
;tree is built of 2-byte nodes
;1st byte !=0: branch node, forward distance -1 to next level if current bit is 0, add two if current bit is 1
;         ==0: leaf node
;2nd byte    : leaf node value/code entry
;input:	YH:YL		start address of Huffman decoding tree to be built
;output: YH:YL		next available address for Huffman decoding tree
build_ht: rd_byte								;get Huffman decoding tree information
	sbiw r25:r24,1								;segment length count -1
	ldi r17,$10
	and r17,r16									;bit 4 carries type of Huffman decoding tree, 0: DC table, 1: AC table
	swap r17
	andi r16,15									;bit 0...3 carry number of Huffman decoding tree, only 0 or 1 supported
	add r16,r16									;construct pointer table index
	add r16,r17
	add r16,r16
	ldi XL,LOW(ht_ptr)
	ldi XH,HIGH(ht_ptr)
	add XL,r16
	adc XH,r6
	st X+,YL									;store start address of current Huffman decoding tree in pointer table
	st X,YH
	movw r5:r4,YH:YL
	ldi XL,LOW(tmp_buf)
	ldi XH,HIGH(tmp_buf)
	ldi r17,16
get_symcnt: rd_byte								;get symbol count for corresponding code length
	st X+,r16
	dec r17
	brne get_symcnt
	ldi r17,16
push_symcnt: ld r16,-X
	push r16									;push symbol count onto stack
	dec r17
	brne push_symcnt
	sbiw r25:r24,16								;subtract 16 from segment length count
	ldi r19,16									;code word level count
	st Y,r6										;create root node
	adiw Y,2									;new end of Huffman decoding tree
	clr r22										;r23:r22 current code word
	clr r23
	ldi r20,1									;r21:r20 current code bit mask
	clr r21
bld_htlp: pop r17								;get symbol count for current code word level
	tst r17
	breq bld_htn								;skip if no symbols at this level
bld_htlp1: rd_byte								;get code value
	sbiw r25:r24,1								;segment length count -1
	rcall add_node								;add node(s) and store code value into leaf node
	subi r22,-1									;increment code word
	sbci r23,-1
	dec r17										;decrement symbol count
	brne bld_htlp1
bld_htn:
	add r22,r22									;advance to next code word level (up to 16 bit)
	adc r23,r23
	add r20,r20
	adc r21,r21
	dec r19										;decrement level count
	brne bld_htlp
	ret

;add new node(s) to Huffman decoding tree if not already existing and store code value into leaf node
;input:	r5:r4		start address of current Huffman decoding tree
;		r16			code value
;		r21:r20		code bit mask
;		r23:r22		current code word
;		YH:YL		next available address after current Huffman decoding tree
add_node: 
	movw r1:r0,r21:r20							;r1:r0 holds current code mask upper bit
	movw XH:XL,r5:r4							;XH:XL points to start address of current Huffman decoding tree
add_node_lp: ld r2,X							;get distance to next level
	tst r2
	brne add_node1								;skip if distance already set
	mov r2,YL
	sub r2,XL									;calculate distance to current end of Huffman decoding tree
	dec r2
	st X,r2										;store distance on current node
	st Y,r6										;create new node for code bit 0
	std Y+2,r6									;create new node for code bit 1
	adiw Y,4									;adapt pointer to end of current Huffman decoding tree
add_node1: movw r3:r2,r1:r0						;get current code bit mask
	and r2,r22									;mask out other bits of current code word
	and r3,r23
	or r2,r3									;check if current bit in code word is set
	ld r18,X+									;get distance to next level
	breq add_node2
	subi r18,-2									;add two if bit in code word was set
add_node2: add XL,r18							;add distance to pointer
	adc XH,r6
	lsr r1										;shift code bit mask 
	ror r0
	brcc add_node_lp							;repeat until all bits were checked
	adiw X,1
	st X,r16									;store code value
	ret

;Inverse Discrete Cosine Transform
;implementation based on Tiny JPEG Decompressor
;input:	YH:YL		pointer to component block storage

.equ m13=181	;32768*cos(4*pi/16),					23170>>7=181
.equ m2=139		;32768*(cos(2*pi/16)-cos(6*pi/16)),		17733>>7=139
.equ m4=167		;32768*(cos(2*pi/16)+cos(6*pi/16)),		42813>>8=167
.equ m5=237		;32768*cos(2*pi/16),					30274>>7=237
idct: movw ZH:ZL,YH:YL
	subi ZL,LOW(-64)
	sbci ZH,HIGH(-64)
	push YL
	push YH

	ldi r19,8

idct_col_lp:
	ld r8,Y					;v0
	ldd r9,Y+1
	ldd r10,Y+32			;v1
	ldd r11,Y+33
	ld r12,Z				;v2
	ldd r13,Z+1
	ldd r14,Z+32			;v3
	ldd r15,Z+33

	movw r17:r16,r11:r10	;t11=(v1-v3)*m13
	sub r16,r14
	sbc r17,r15
 	ldi r18,m13
	clr r2
	fmul r16,r18
	adc r2,r2
	mov r3,r1
	fmulsu r17,r18
	add r0,r3
	adc r1,r2

	movw r27:r26,r9:r8		;t10=v0+v2
	add r26,r12
	adc r27,r13

	movw r25:r24,r9:r8		;t12=v0-v2
	sub r24,r12
	sbc r25,r13

	add r14,r10				;v3+=v1
	adc r15,r11

	sub r0,r14				;t11-=v3
	sbc r1,r15

	movw r9:r8,r27:r26		;v0=t10+v3
	add r8,r14
	adc r9,r15
	
	sub r26,r14				;(v3)=t10-v3
	sbc r27,r15

	movw r23:r22,r1:r0		;v1=t11+t12
	add r22,r24
	adc r23,r25

	sub r24,r0				;(v2)=t12-t11
	sbc r25,r1

	ldd r12,Z+48			;v4
	ldd r13,Z+49
	ldd r14,Y+16			;v5
	ldd r15,Y+17

	movw r7:r6,r15:r14		;t10=v5-v4
	sub r6,r12
	sbc r7,r13

	movw r5:r4,r15:r14		;t11=v5+v4
	add r4,r12
	adc r5,r13

	ldd r12,Z+16			;v6
	ldd r13,Z+17
	ldd r14,Y+48			;v7
	ldd r15,Y+49

	movw r21:r20,r13:r12	;t12=v6-v7
	sub r20,r14
	sbc r21,r15

	add r14,r12				;v7+=v6
	adc r15,r13

	movw r17:r16,r5:r4		;v5=(t11-v7)*m13
	sub r16,r14
	sbc r17,r15
	ldi r18,m13
	clr r11
	fmul r16,r18
	adc r11,r11
	mov r10,r1
	fmulsu r17,r18
	add r10,r0
	adc r11,r1

	add r14,r4				;v7+=t11
	adc r15,r5

	movw r17:r16,r7:r6		;t13=(t10+t12)*m5
	add r16,r20
	adc r17,r21
	ldi r18,m5
	clr r5
	fmul r16,r18
	adc r5,r5
	mov r4,r1
	fmulsu r17,r18
	add r4,r0
	adc r5,r1

	movw r17:r16,r7:r6		;v4=t13-t10*m2
	ldi r18,m2
	clr r2
	fmul r16,r18
	adc r2,r2
	mov r3,r1
	fmulsu r17,r18
	add r0,r3
	adc r1,r2
	movw r13:r12,r5:r4
	sub r12,r0
	sbc r13,r1

	ldi r18,m4				;(v6)=t13-t12*m4-v7
	clr r2
	fmul r20,r18
	adc r2,r2
	mov r20,r0
	mov r3,r1
	fmulsu r21,r18
	add r0,r3
	adc r1,r2
	add r20,r20
	adc r0,r0
	adc r1,r1
	sub r4,r0
	sbc r5,r1
	sub r4,r14
	sbc r5,r15

	sub r10,r4				;v5-=v6
	sbc r11,r5

	sub r12,r10				;v4-=v5
	sbc r13,r11

	movw r1:r0,r9:r8		;v0+v7
	add r0,r14
	adc r1,r15
	st Y+,r0
	st Y+,r1

	sub r8,r14				;v0-v7
	sbc r9,r15
	std Z+48,r8
	std Z+49,r9

	movw r1:r0,r23:r22		;v1+v6
	add r0,r4
	adc r1,r5
	std Y+14,r0
	std Y+15,r1

	sub r22,r4				;v1-v6
	sbc r23,r5
	std Z+32,r22
	std Z+33,r23

	movw r1:r0,r25:r24		;v2+v5
	add r0,r10
	adc r1,r11
	std Y+30,r0
	std Y+31,r1

	sub r24,r10				;v2+v5
	sbc r25,r11
	std Z+16,r24
	std Z+17,r25

	movw r1:r0,r27:r26		;v3+v4
	add r0,r12
	adc r1,r13
	std Y+46,r0
	std Y+47,r1

	sub r26,r12				;v3-v4
	sbc r27,r13
	st Z+,r26
	st Z+,r27

	dec r19
	breq idct1
	rjmp idct_col_lp
idct1:

	pop YH
	pop YL
	ldi r19,8

idct_row_lp:
	ld r20,Y				;v0
	ldd r21,Y+1
	ldd r10,Y+4				;v1
	ldd r11,Y+5
	ldd r12,Y+8				;v2
	ldd r13,Y+9
	ldd r14,Y+12			;v3
	ldd r15,Y+13

	subi r20,LOW(-$400)		;v0+=-$400
	sbci r21,HIGH(-$400)

	movw r17:r16,r11:r10	;t11=(v1-v3)*m13
	sub r16,r14
	sbc r17,r15
 	ldi r18,m13
	clr r2
	fmul r16,r18
	adc r2,r2
	mov r3,r1
	fmulsu r17,r18
	add r0,r3
	adc r1,r2

	movw r27:r26,r21:r20	;t10=v0+v2
	add r26,r12
	adc r27,r13

	movw r25:r24,r21:r20	;t12=v0-v2
	sub r24,r12
	sbc r25,r13

	add r14,r10				;v3+=v1
	adc r15,r11

	sub r0,r14				;t11-=v3
	sbc r1,r15

	movw r31:r30,r27:r26	;v0=t10+v3
	add r30,r14
	adc r31,r15
	
	sub r26,r14				;(v3)=t10-v3
	sbc r27,r15

	movw r23:r22,r1:r0		;v1=t11+t12
	add r22,r24
	adc r23,r25

	sub r24,r0				;(v2)=t12-t11
	sbc r25,r1

	ldd r8,Y+14				;v4
	ldd r9,Y+15
	ldd r10,Y+2				;v5
	ldd r11,Y+3
	ldd r12,Y+10			;v6
	ldd r13,Y+11
	ldd r14,Y+6				;v7
	ldd r15,Y+7

	movw r7:r6,r11:r10		;t10=v5-v4
	sub r6,r8
	sbc r7,r9

	movw r5:r4,r11:r10		;t11=v5+v4
	add r4,r8
	adc r5,r9

	movw r21:r20,r13:r12	;t12=v6-v7
	sub r20,r14
	sbc r21,r15

	add r14,r12				;v7+=v6
	adc r15,r13

	movw r17:r16,r5:r4		;v5=(t11-v7)*m13
	sub r16,r14
	sbc r17,r15
	ldi r18,m13
	clr r11
	fmul r16,r18
	adc r11,r11
	mov r10,r1
	fmulsu r17,r18
	add r10,r0
	adc r11,r1

	add r14,r4				;v7+=t11
	adc r15,r5

	movw r17:r16,r7:r6		;t13=(t10+t12)*m5
	add r16,r20
	adc r17,r21
	ldi r18,m5
	clr r5
	fmul r16,r18
	adc r5,r5
	mov r4,r1
	fmulsu r17,r18
	add r4,r0
	adc r5,r1

	movw r17:r16,r7:r6		;v4=t13-t10*m2
	ldi r18,m2
	clr r2
	fmul r16,r18
	adc r2,r2
	mov r3,r1
	fmulsu r17,r18
	add r0,r3
	adc r1,r2
	movw r9:r8,r5:r4
	sub r8,r0
	sbc r9,r1

	ldi r18,m4				;(v6)=t13-t12*m4-v7
	clr r2
	fmul r20,r18
	adc r2,r2
	mov r20,r0
	mov r3,r1
	fmulsu r21,r18
	add r0,r3
	adc r1,r2
	add r20,r20
	adc r0,r0
	adc r1,r1
	sub r4,r0
	sbc r5,r1
	sub r4,r14
	sbc r5,r15

	sub r10,r4				;v5-=v6
	sbc r11,r5

	sub r8,r10				;v4-=v5
	sbc r9,r11

	.macro byteclip 
	brpl bc1
	clr @0
	rjmp bc3
bc1: lsr @1
	ror @0
	lsr @1
	ror @0
	lsr @1
	breq bc2
	ldi @0,-1
	rjmp bc3
bc2: ror @0
bc3:
	.endmacro

	movw r21:r20,r31:r30	;v0+v7
	add r20,r14
	adc r21,r15
	byteclip r20,r21
	st Y,r20

	sub r30,r14				;v0-v7
	sbc r31,r15
	byteclip r30,r31
	std Y+14,r30

	movw r21:r20,r23:r22	;v1+v6
	add r20,r4
	adc r21,r5
	byteclip r20,r21
	std Y+2,r20

	sub r22,r4				;v1-v6
	sbc r23,r5
	byteclip r22,r23
	std Y+12,r22

	movw r21:r20,r25:r24	;v2+v5
	add r20,r10
	adc r21,r11
	byteclip r20,r21
	std Y+4,r20

	sub r24,r10				;v2-v5
	sbc r25,r11
	byteclip r24,r25
	std Y+10,r24

	movw r21:r20,r27:r26	;v3+v4
	add r20,r8
	adc r21,r9
	byteclip r20,r21
	std Y+6,r20

	sub r26,r8				;v3-v4
	sbc r27,r9
	byteclip r26,r27
	std Y+8,r26

	adiw Y,16
	dec r19
	breq idct_x
	rjmp idct_row_lp
idct_x: ret

;set edges for drawing area on LCD
;input:	r1:r0		upper/left edge
;		r21			LCD command for particular edge definition
;		XH:XL		lower/right edge
lcd_set_edge:
	mov r18,r16
	in r16,DISPLAYPORT
	andi r16,~(1<<DISPLAY_SS|1<<DISPLAY_C_D)	;assert /SS and set LCD command mode
	out DISPLAYPORT,r16
	mov r16,r21
	rcall lcd_wr_spi
	rcall lcd_spi_wt
	sbi DISPLAYPORT,DISPLAY_C_D					;set LCD data mode
	mov r16,r1				;X/YSH
	rcall lcd_wr_spi
	mov r16,r0				;X/YSL
	rcall lcd_wr_spi
	mov r16,XH				;X/YEH
	rcall lcd_wr_spi
	mov r16,XL				;X/YEL
	rcall lcd_wr_spi
	rcall lcd_spi_wt
	sbi DISPLAYPORT,DISPLAY_SS					;release /SS
	mov r16,r18
	ret

;start to send current line/double line to LCD
;input:	r23:r22		current buffer address
;output: r23:r22	next buffer address
	.if lcd_dma==1
lcd_send_buf:
dma_wait:
	lds r16,(DMA_STATUS)						;wait if DMA is still busy
	sbrc r16,DMA_CH0BUSY_bp
	rjmp dma_wait	

	rcall lcd_spi_wt							;wait until final byte was transmitted to LCD
	sts (DISPLAY_IF+USART_STATUS_offset),r16	;clear USART TXC flag

	sts (DMA_CH0_SRCADDR0),r22					;reload current buffer address
	sts (DMA_CH0_SRCADDR1),r23
	sts (DMA_CH0_SRCADDR2),r3
	ldi r16,DMA_CH_ENABLE_bm|DMA_CH_SINGLE_bm	;start DMA
	sts (DMA_CH0_CTRLA),r16
	cpi r22,LOW(tmp_buf)						;double buffering
	breq lcd_send_buf1
	subi r22,96									;use lower buffer for next line
	sbci r23,0
	ret
lcd_send_buf1: subi r22,-96						;use higher buffer for next line
	sbci r23,-1
	ret
	.else
;using no DMA is noticeably slower
;interrupt driven transmission might be an option, but the interrupt handling overhead has to be considered
lcd_send_buf:
	movw ZH:ZL,r23:r22
	mov r8,r9
lcd_send_buf_lp: ld r16,Z+
	rcall lcd_wr_spi
	dec r8
	brne lcd_send_buf_lp
	ret											;no double buffering
	.endif

;delay
;input: r16			delay in milliseconds
wait: sts (tmcnt),r16
wait_lp: lds r16,(tmcnt)
	tst r16
	brne wait_lp
	ret

;wait until current SPI transmission to LCD is finished
lcd_spi_wt:
	lds r16,(DISPLAY_IF+USART_STATUS_offset)
	andi r16,USART_TXCIF_bm
	breq lcd_spi_wt
	ret

;transmit byte to LCD
;input: r16			byte to send
lcd_wr_spi: push r16
	rcall lcd_spi_wt
	sts (DISPLAY_IF+USART_STATUS_offset),r16
	pop r16
	sts (DISPLAY_IF+USART_DATA_offset),r16
	ret

	.if flash_rd==1
;get single byte from controller's flash memory
;capable to handle memory addresses >65535
;output: r16		byte read from flash
rd_byte_flash:
	push ZL
	push ZH
	.if big_flash==1
	push r0
	in r0,CPU_RAMPZ
	lds ZL,(rd_ptr+2)
	out CPU_RAMPZ,ZL
	.endif
	lds ZL,(rd_ptr)
	lds ZH,(rd_ptr+1)
	.if big_flash==1
	elpm r16,Z+
	.else
	lpm r16,Z+
	.endif
	sts (rd_ptr),ZL
	sts (rd_ptr+1),ZH
	.if big_flash==1
	in ZL,CPU_RAMPZ
	sts (rd_ptr+2),ZL
	out CPU_RAMPZ,r0
	pop r0
	.endif
	pop ZH
	pop ZL
	ret
	.endif

	.if usart_rd
;get single byte from serial interface's ring buffer
;handle RTS signal
;output: r16		byte read from serial receive buffer
rd_byte_sbuf: push ZL
	push ZH
rd_byte_wt: lds ZL,(rd_sptr)
	lds ZH,(wr_sptr)
	cp ZL,ZH
	breq rd_byte_wt
	sbis RTS
	rjmp rd_byte_sbuf1
	sub ZH,ZL
	cpi ZH,240
	brcc rd_byte_sbuf1
	cbi RTS
rd_byte_sbuf1: clr ZH
	subi ZL,LOW(-serbuf)
	sbci ZH,HIGH(-serbuf)
	ld r16,Z+
	subi ZL,LOW(serbuf)
	sts (rd_sptr),ZL
	pop ZH
	pop ZL
	ret 
	.endif

	.if spiram_rd==1
;wait until current SPI transmission over external RAM interface is finished
	.if RAM_IF_MODE==1
;USART in master SPI mode
ram_spi_wt:
	lds r16,(RAM_IF+USART_STATUS_offset)
	andi r16,USART_TXCIF_bm
	breq ram_spi_wt
	ret
	.else
;native SPI
ram_spi_wt:
	lds r16,(RAM_IF+SPI_STATUS_offset)
	andi r16,SPI_IF_bm
	breq ram_spi_wt
	ret
	.endif

;set address pointer for external RAM
;input:	r20:r19:r18	current address
set_rampos: cbi RAMPORT,RAMPORT_SS				;assert /SS for external RAM
	ldi r16,3									;serial RAM 23LC1024 read command
	rcall ram_wr_rd_spi
	mov r16,r20
	rcall ram_wr_rd_spi
	mov r16,r19
	rcall ram_wr_rd_spi
	mov r16,r18

;transmit/read byte to/from external RAM
;input:	r16		byte to store in external RAM
;output: r16	byte read from external RAM
	.if RAM_IF_MODE==1
;USART in master SPI mode
ram_wr_rd_spi: sts (RAM_IF+USART_DATA_offset),r16
	rcall ram_spi_wt
	sts (RAM_IF+USART_STATUS_offset),r16
	lds r16,(RAM_IF+USART_DATA_offset)
	ret
	.else
;native SPI
ram_wr_rd_spi: sts (RAM_IF+SPI_DATA_offset),r16
	rcall ram_spi_wt
	lds r16,(RAM_IF+SPI_DATA_offset)
	ret
	.endif

	.endif

	.if spiram_rd==1
;minimal loader to store precompiled image data received via serial interface into external RAM
load_ram:
	ldi YL,LOW(SERIAL_IF)						;set up serial interface
	ldi YH,HIGH(SERIAL_IF)
	ldi r16,LOW(bsel)
	std Y+USART_BAUDCTRLA_offset,r16
	ldi r16,bscale<<4|HIGH(bsel)
	std Y+USART_BAUDCTRLB_offset,r16
	ldi r16,USART_CMODE_ASYNCHRONOUS_gc|USART_CHSIZE_8BIT_gc	;async 8n1
	std Y+USART_CTRLC_offset,r16
	ldi r16,USART_RXEN_bm|USART_TXEN_bm
	std Y+USART_CTRLB_offset,r16
	ldi r16,USART_RXCIF_bm
	std Y+USART_STATUS_offset,r16

	ldi ZL,LOW(ld_prompt*2)
	ldi ZH,HIGH(ld_prompt*2)
	rcall send_str								;prompt user to start sending the data

	cbi RAMPORT,RAMPORT_SS						;assert /SS for external RAM
	ldi r16,2									;serial RAM 23LC1024 write command
	rcall ram_wr_rd_spi
	clr r16										;set address in external RAM to 0
	rcall ram_wr_rd_spi
	clr r16
	rcall ram_wr_rd_spi
	clr r16
	rcall ram_wr_rd_spi

recv_lp: clr r0
	clr r1
	ldi r18,LOW(clock*4)						;define timeout to determine if data reception has finished
recv_wt: ldd r16,Y+USART_STATUS_offset
 	andi r16,USART_RXCIF_bm
 	brne recv1
 	dec r0										;decrement timeout counter
	brne recv_wt
	dec r1
	brne recv_wt
	dec r18
	brne recv_wt
	sbi RAMPORT,RAMPORT_SS						;/SS for SPI RAM to HIGH
 	ldi ZL,LOW(ld_fin*2)
	ldi ZH,HIGH(ld_fin*2)
	rcall send_str								;inform user that data reception has been finished
	.if RAM_IF_MODE==1
	ldi r16,-1
	sts (RAM_IF+USART_STATUS_offset),r16		;clear status of interface to external RAM
	lds r16,(RAM_IF+USART_DATA_offset)			;empty receive buffer
	.else
	lds r16,(RAM_IF+SPI_STATUS_offset)			;clear status of interface to external RAM
	lds r16,(RAM_IF+SPI_DATA_offset)			;empty receive buffer
	.endif
	rjmp restart
recv1: ldd r16,Y+USART_DATA_offset				;read byte from serial interface
	.if RAM_IF_MODE==1
;USART in master SPI mode
	sts (RAM_IF+USART_DATA_offset),r16			;and store it in external RAM, no status check since RAM interface is clocked faster than speed of serial interface
	.else
;native SPI
	sts (RAM_IF+SPI_DATA_offset),r16
	.endif
	rjmp recv_lp

;send zero-terminated string over serial interface
;input:	ZH:ZL		pointer to string data
send_str: lpm r16,Z+
	tst r16
	brne send_str1
	ret
send_str1: rcall send_chr
	rjmp send_str

;dump hexadecimal byte over serial interface
;input:	r16			byte to dump
ser_dump: push r16
	swap r16
	rcall ser_dmp_nb
	pop r16
ser_dmp_nb: andi r16,15
	subi r16,-'0'
	cpi r16,$3a
	brcs send_chr
	subi r16,-7

;send single character over serial interface
;input:	r16			byte to send
send_chr: ldd r17,Y+USART_STATUS_offset
	sbrs r17,USART_DREIF_bp
	rjmp send_chr
	std Y+USART_DATA_offset,r16
	ret
	.endif

	.org (PC+$ff) & $ffff00
;tables needed for conversion to RGB
rgb_cr_tbl1_402: .dw -179,-178,-176,-175,-173,-172,-171,-169,-168,-166,-165,-164,-162,-161,-159,-158,-157,-155,-154,-152,-151,-150,-148,-147,-145,-144,-143,-141,-140,-138,-137,-135,-134,-133,-131,-130,-128,-127,-126,-124,-123,-121,-120,-119,-117,-116,-114,-113,-112,-110,-109,-107,-106,-105,-103,-102,-100,-99,-98,-96,-95,-93,-92,-91,-89,-88,-86,-85,-84,-82,-81,-79,-78,-77,-75,-74,-72,-71,-70,-68,-67,-65,-64,-63,-61,-60,-58,-57,-56,-54,-53,-51,-50,-49,-47,-46,-44,-43,-42,-40,-39,-37,-36,-35,-33,-32,-30,-29,-28,-26,-25,-23,-22,-21,-19,-18,-16,-15,-14,-12,-11,-9,-8,-7,-5,-4,-2,-1,0,1,2,4,5,7,8,9,11,12,14,15,16,18,19,21,22,23,25,26,28,29,30,32,33,35,36,37,39,40,42,43,44,46,47,49,50,51,53,54,56,57,58,60,61,63,64,65,67,68,70,71,72,74,75,77,78,79,81,82,84,85,86,88,89,91,92,93,95,96,98,99,100,102,103,105,106,107,109,110,112,113,114,116,117,119,120,121,123,124,126,127,128,130,131,133,134,135,137,138,140,141,143,144,145,147,148,150,151,152,154,155,157,158,159,161,162,164,165,166,168,169,171,172,173,175,176,178
rgb_cb_tbl1_772: .dw -226,-225,-223,-221,-219,-217,-216,-214,-212,-210,-209,-207,-205,-203,-202,-200,-198,-196,-194,-193,-191,-189,-187,-186,-184,-182,-180,-178,-177,-175,-173,-171,-170,-168,-166,-164,-163,-161,-159,-157,-155,-154,-152,-150,-148,-147,-145,-143,-141,-139,-138,-136,-134,-132,-131,-129,-127,-125,-124,-122,-120,-118,-116,-115,-113,-111,-109,-108,-106,-104,-102,-101,-99,-97,-95,-93,-92,-90,-88,-86,-85,-83,-81,-79,-77,-76,-74,-72,-70,-69,-67,-65,-63,-62,-60,-58,-56,-54,-53,-51,-49,-47,-46,-44,-42,-40,-38,-37,-35,-33,-31,-30,-28,-26,-24,-23,-21,-19,-17,-15,-14,-12,-10,-8,-7,-5,-3,-1,0,1,3,5,7,8,10,12,14,15,17,19,21,23,24,26,28,30,31,33,35,37,38,40,42,44,46,47,49,51,53,54,56,58,60,62,63,65,67,69,70,72,74,76,77,79,81,83,85,86,88,90,92,93,95,97,99,101,102,104,106,108,109,111,113,115,116,118,120,122,124,125,127,129,131,132,134,136,138,139,141,143,145,147,148,150,152,154,155,157,159,161,163,164,166,168,170,171,173,175,177,178,180,182,184,186,187,189,191,193,194,196,198,200,202,203,205,207,209,210,212,214,216,217,219,221,223,225
rgb_cr_tbl0_714: .db 91,90,89,89,88,87,87,86,85,84,84,83,82,82,81,80,79,79,78,77,77,76,75,74,74,73,72,72,71,70,69,69,68,67,67,66,65,64,64,63,62,62,61,60,59,59,58,57,57,56,55,54,54,53,52,52,51,50,49,49,48,47,47,46,45,44,44,43,42,42,41,40,39,39,38,37,37,36,35,34,34,33,32,32,31,30,29,29,28,27,27,26,25,24,24,23,22,22,21,20,19,19,18,17,17,16,15,14,14,13,12,12,11,10,9,9,8,7,7,6,5,4,4,3,2,2,1,0,0,-0,-1,-2,-2,-3,-4,-4,-5,-6,-7,-7,-8,-9,-9,-10,-11,-12,-12,-13,-14,-14,-15,-16,-17,-17,-18,-19,-19,-20,-21,-22,-22,-23,-24,-24,-25,-26,-27,-27,-28,-29,-29,-30,-31,-32,-32,-33,-34,-34,-35,-36,-37,-37,-38,-39,-39,-40,-41,-42,-42,-43,-44,-44,-45,-46,-47,-47,-48,-49,-49,-50,-51,-52,-52,-53,-54,-54,-55,-56,-57,-57,-58,-59,-59,-60,-61,-62,-62,-63,-64,-64,-65,-66,-67,-67,-68,-69,-69,-70,-71,-72,-72,-73,-74,-74,-75,-76,-77,-77,-78,-79,-79,-80,-81,-82,-82,-83,-84,-84,-85,-86,-87,-87,-88,-89,-89,-90
rgb_cb_tbl0_344: .db 44,43,43,43,42,42,41,41,41,40,40,40,39,39,39,38,38,38,37,37,37,36,36,36,35,35,35,34,34,34,33,33,33,32,32,31,31,31,30,30,30,29,29,29,28,28,28,27,27,27,26,26,26,25,25,25,24,24,24,23,23,23,22,22,22,21,21,20,20,20,19,19,19,18,18,18,17,17,17,16,16,16,15,15,15,14,14,14,13,13,13,12,12,12,11,11,11,10,10,9,9,9,8,8,8,7,7,7,6,6,6,5,5,5,4,4,4,3,3,3,2,2,2,1,1,1,0,0,0,-0,-0,-1,-1,-1,-2,-2,-2,-3,-3,-3,-4,-4,-4,-5,-5,-5,-6,-6,-6,-7,-7,-7,-8,-8,-8,-9,-9,-9,-10,-10,-11,-11,-11,-12,-12,-12,-13,-13,-13,-14,-14,-14,-15,-15,-15,-16,-16,-16,-17,-17,-17,-18,-18,-18,-19,-19,-19,-20,-20,-20,-21,-21,-22,-22,-22,-23,-23,-23,-24,-24,-24,-25,-25,-25,-26,-26,-26,-27,-27,-27,-28,-28,-28,-29,-29,-29,-30,-30,-30,-31,-31,-31,-32,-32,-33,-33,-33,-34,-34,-34,-35,-35,-35,-36,-36,-36,-37,-37,-37,-38,-38,-38,-39,-39,-39,-40,-40,-40,-41,-41,-41,-42,-42,-43,-43,-43

;table needed for prescaling according to Arai's IDCT algorithm in zigzag order
arai_scale_zz:
.dw int(1.38704*32768) ;1
.dw int(1.38704*32768) ;2
.dw int(1.30656*32768) ;3
.dw int(1.92388*32768) ;4
.dw int(1.30656*32768) ;5
.dw int(1.17588*32768) ;6
.dw int(1.81226*32768) ;7
.dw int(1.81226*32768) ;8
.dw int(1.17588*32768) ;9
.dw int(1.00000*32768) ;10
.dw int(1.63099*32768) ;11
.dw int(1.70711*32768) ;12
.dw int(1.63099*32768) ;13
.dw int(1.00000*32768) ;14
.dw int(0.78570*32768) ;15
.dw int(1.38704*32768) ;16
.dw int(1.53636*32768) ;17
.dw int(1.53636*32768) ;18
.dw int(1.38704*32768) ;19
.dw int(0.78570*32768) ;20
.dw int(0.54120*32768) ;21
.dw int(1.08979*32768) ;22
.dw int(1.30656*32768) ;23
.dw int(1.38268*32768) ;24
.dw int(1.30656*32768) ;25
.dw int(1.08979*32768) ;26
.dw int(0.54120*32768) ;27
.dw int(0.27590*32768) ;28
.dw int(0.75066*32768) ;29
.dw int(1.02656*32768) ;30
.dw int(1.17588*32768) ;31
.dw int(1.17588*32768) ;32
.dw int(1.02656*32768) ;33
.dw int(0.75066*32768) ;34
.dw int(0.27590*32768) ;35
.dw int(0.38268*32768) ;36
.dw int(0.70711*32768) ;37
.dw int(0.92388*32768) ;38
.dw int(1.00000*32768) ;39
.dw int(0.92388*32768) ;40
.dw int(0.70711*32768) ;41
.dw int(0.38268*32768) ;42
.dw int(0.36048*32768) ;43
.dw int(0.63638*32768) ;44
.dw int(0.78570*32768) ;45
.dw int(0.78570*32768) ;46
.dw int(0.63638*32768) ;47
.dw int(0.36048*32768) ;48
.dw int(0.32442*32768) ;49
.dw int(0.54120*32768) ;50
.dw int(0.61732*32768) ;51
.dw int(0.54120*32768) ;52
.dw int(0.32442*32768) ;53
.dw int(0.27590*32768) ;54
.dw int(0.42522*32768) ;55
.dw int(0.42522*32768) ;56
.dw int(0.27590*32768) ;57
.dw int(0.21678*32768) ;58
.dw int(0.29290*32768) ;59
.dw int(0.21677*32768) ;60
.dw int(0.14932*32768) ;61
.dw int(0.14932*32768) ;62
.dw int(0.07612*32768) ;63

;table with power of 2 values
pow2tab: .dw 1,2,4,8,16,32,64,128,256,512,1024

;LCD initialization values (ILI9341 controller), especially positive and negative Gamma correction
;coding: 1 byte  LCD command (0: no command, end of string)
;		 1 byte  parameter length (0: no parameter, -1: no parameter but 10ms delay to the following command)
;		 n bytes parameter(s)
#ifdef LCD_CTRL_ILI9341
disp_init_str: .db $11,-1,$29,0,$36,1,0b10001000,$e0,15,$0F,$31,$2B,$0C,$0E,$08,$4E,$F1,$37,$07,$10,$03,$0E,$09,$00,$e1,15,$00,$0E,$14,$03,$11,$07,$31,$C1,$48,$08,$0F,$0C,$31,$36,$0F,0
#endif
;LCD initialization values (ST7735R controller), Gamma Curve 2 (G1,8) selected
#ifdef LCD_CTRL_ST7735R
disp_init_str:	.db $11,-1,$29,0,$26,1,2,0 
;including Gamma correction, didn't have effect with my display
;disp_init_str:	.db $11,-1,$29,0,$26,1,2,$e0,16,$09,$16,$09,$20,$21,$1B,$13,$19,$17,$15,$1E,$2B,$04,$05,$02,$0E,$e1,16,$0B,$14,$08,$1E,$22,$1D,$18,$1E,$1B,$1A,$24,$2B,$06,$06,$02,$0F,0
#endif

	.if spiram_rd==1
ld_prompt: .db 13,10,"Start to send data in BINARY mode...",13,10,0
ld_fin: .db "Loading finished",13,10,0
	.endif

	.if flash_rd==1
/*
files: .dd 22<<24|f1*2,22<<24|f2*2,22<<24|f3*2,0	;MSB not needed for addressing, used to code delay (x^2) between two pictures; address=0 jumps to beginning of list
f1: .include "Flowers 320x240\1.inc"
f2: .include "Flowers 320x240\2.inc"
f3: .include "Flowers 320x240\3.inc"
*/
/*
files: .dd f1*2,f2*2,f3*2,f4*2,f5*2,f6*2,f7*2,f8*2,f9*2,f10*2,f11*2,f12*2,0
f1: .include "Animation 160x128\01.inc"
f2: .include "Animation 160x128\02.inc"
f3: .include "Animation 160x128\03.inc"
f4: .include "Animation 160x128\04.inc"
f5: .include "Animation 160x128\05.inc"
f6: .include "Animation 160x128\06.inc"
f7: .include "Animation 160x128\07.inc"
f8: .include "Animation 160x128\08.inc"
f9: .include "Animation 160x128\09.inc"
f10: .include "Animation 160x128\10.inc"
f11: .include "Animation 160x128\11.inc"
f12: .include "Animation 160x128\12.inc"
*/

files: .dd 22<<24|f1*2,22<<24|f2*2,22<<24|f3*2,22<<24|f4*2,22<<24|f5*2,22<<24|f6*2,22<<24|f7*2,0
f1:	.include "Flowers 160x128\Flowers1.inc"
f2:	.include "Flowers 160x128\Flowers2.inc"
f3:	.include "Flowers 160x128\Flowers3.inc"
f4:	.include "Flowers 160x128\Flowers4.inc"
f5:	.include "Flowers 160x128\Flowers5.inc"
f6:	.include "Flowers 160x128\Flowers6.inc"
f7:	.include "Flowers 160x128\Flowers7.inc"

	.endif

	.dseg
	.org SRAM_START
;memory storage needed for decoding
;the arrangement of the fields must be preserved to make sure address arithmetics will work properly
;must start on a 256 byte boundary
blk_y: .byte 128*4
blk_cb: .byte 128
blk_cr: .byte 128
	.if lcd_dma==1
tmp_buf: .byte 192								;double buffering for LCD DMA, max. 2x96 bytes needed
	.else
tmp_buf: .byte 128								;no double buffering without LCD DMA
	.endif
q_tbl:	.byte 128
prv_dc: .byte 3*2
sfact: .byte 1
tmcnt: .byte 1
ht_ptr: .byte 4*2
 .if flash_rd==1
 .if big_flash==1
rd_ptr: .byte 3
 .else
rd_ptr: .byte 2
 .endif
 .endif
 .if usart_rd==1
wr_sptr: .byte 1
rd_sptr: .byte 1
serbuf: .byte 256
 .endif
 .if spiram_rd==1 || flash_rd==1
fpos: .byte 1
 .endif
ht:	.byte 1400



